/* -*- Mode: C; c-basic-offset:4 ; indent-tabs-mode:nil -*- */
/*
 * Copyright (c) 2014-2015 Intel, Inc.  All rights reserved.
 * Copyright (c) 2014      Research Organization for Information Science
 *                         and Technology (RIST). All rights reserved.
 * Copyright (c) 2014-2015 Artem Y. Polyakov <artpol84@gmail.com>.
 *                         All rights reserved.
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */

#include <private/autogen/config.h>
#include <pmix/rename.h>
#include <private/types.h>
#include <private/pmix_stdint.h>
#include <private/pmix_socket_errno.h>

#include <pmix_server.h>

#include "src/include/pmix_globals.h"

#ifdef HAVE_STRING_H
#include <string.h>
#endif
#include <fcntl.h>
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif
#ifdef HAVE_SYS_SOCKET_H
#include <sys/socket.h>
#endif
#ifdef HAVE_SYS_UN_H
#include <sys/un.h>
#endif
#ifdef HAVE_SYS_UIO_H
#include <sys/uio.h>
#endif
#ifdef HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif
#include <ctype.h>
#include PMIX_EVENT_HEADER

#include "src/util/argv.h"
#include "src/util/error.h"
#include "src/util/output.h"
#include "src/util/pmix_environ.h"
#include "src/util/progress_threads.h"
#include "src/usock/usock.h"
#include "src/sec/pmix_sec.h"

#include "pmix_server_ops.h"

// local variables
static int init_cntr = 0;
static char *myuri = NULL;
static struct sockaddr_un myaddress;
static bool using_internal_comm = false;
static char *security_mode = NULL;

// local functions for connection support
static void server_message_handler(struct pmix_peer_t *pr, pmix_usock_hdr_t *hdr,
                                   pmix_buffer_t *buf, void *cbdata);
// global variables
pmix_globals_t pmix_globals = {
    .evbase = NULL,
    .debug_output = -1,
    .errhandler = NULL,
    .server = true,
};
pmix_server_globals_t pmix_server_globals;

/* queue a message to be sent to one of our procs - must
 * provide the following params:
 *
 * p - the peer object of the process
 * t - tag to be sent to
 * b - buffer to be sent
 */
#define PMIX_SERVER_QUEUE_REPLY(p, t, b)                                \
    do {                                                                \
        pmix_usock_send_t *snd;                                         \
        pmix_output_verbose(2, pmix_globals.debug_output,               \
                            "[%s:%d] queue reply to %s:%d on tag %d",   \
                            __FILE__, __LINE__,                         \
                            (p)->info->nptr->nspace,                    \
                            (p)->info->rank, (t));                      \
        snd = PMIX_NEW(pmix_usock_send_t);                              \
        snd->hdr.pindex = pmix_globals.pindex;                          \
        snd->hdr.tag = (t);                                             \
        snd->hdr.nbytes = (b)->bytes_used;                              \
        snd->data = (b);                                                \
        /* always start with the header */                              \
        snd->sdptr = (char*)&snd->hdr;                                  \
        snd->sdbytes = sizeof(pmix_usock_hdr_t);                        \
                                                                        \
        /* if there is no message on-deck, put this one there */        \
        if (NULL == (p)->send_msg) {                                    \
            (p)->send_msg = snd;                                        \
        } else {                                                        \
            /* add it to the queue */                                   \
            pmix_list_append(&(p)->send_queue, &snd->super);            \
        }                                                               \
        /* ensure the send event is active */                           \
        if (!(p)->send_ev_active) {                                     \
            event_add(&(p)->send_event, 0);                             \
            (p)->send_ev_active = true;                                 \
        }                                                               \
    } while(0);

static pmix_status_t initialize_server_base(pmix_server_module_t *module)
{
    int debug_level;
    pid_t pid;
    char *tdir, *evar;

    /* initialize the output system */
    if (!pmix_output_init()) {
        return PMIX_ERR_INIT;
    }
    /* Zero globals */
    memset(&pmix_globals, 0, sizeof(pmix_globals));
    memset(&pmix_server_globals, 0, sizeof(pmix_server_globals));
    pmix_server_globals.listen_socket = -1;
    
    /* setup the globals */
    (void)strncpy(pmix_globals.nspace, "pmix-server", PMIX_MAX_NSLEN);
    pmix_globals.debug_output = -1;
    PMIX_CONSTRUCT(&pmix_server_globals.nspaces, pmix_list_t);
    PMIX_CONSTRUCT(&pmix_server_globals.clients, pmix_pointer_array_t);
    pmix_pointer_array_init(&pmix_server_globals.clients, 1, INT_MAX, 1);

    PMIX_CONSTRUCT(&pmix_server_globals.collectives, pmix_list_t);
    PMIX_CONSTRUCT(&pmix_server_globals.dmodex, pmix_list_t);
    PMIX_CONSTRUCT(&pmix_server_globals.localmodex, pmix_list_t);

    /* see if debug is requested */
    if (NULL != (evar = getenv("PMIX_DEBUG"))) {
        debug_level = strtol(evar, NULL, 10);
        pmix_globals.debug_output = pmix_output_open(NULL);
        pmix_output_set_verbosity(pmix_globals.debug_output, debug_level);
    }

    pmix_output_verbose(2, pmix_globals.debug_output,
                        "pmix:server init called");

    /* setup the function pointers */
    memset(&pmix_host_server, 0, sizeof(pmix_server_module_t));
    pmix_host_server = *module;

    /* initialize the datatype support */
    pmix_bfrop_open();

    /* init security */
    pmix_sec_init();
    security_mode = strdup(pmix_sec.name);
    
    /* setup the path to the daemon rendezvous point, using our
     * pid as the "rank" */
    pid = getpid();

    /* find the temp dir */
    if (NULL == (tdir = getenv("TMPDIR"))) {
        if (NULL == (tdir = getenv("TEMP"))) {
            if (NULL == (tdir = getenv("TMP"))) {
                tdir = "/tmp";
            }
        }
    }
    /* now set the address */
    memset(&myaddress, 0, sizeof(struct sockaddr_un));
    myaddress.sun_family = AF_UNIX;
    snprintf(myaddress.sun_path, sizeof(myaddress.sun_path)-1, "%s/pmix-%d", tdir, pid);
    asprintf(&myuri, "%lu:%s", (unsigned long)pid, myaddress.sun_path);


    pmix_output_verbose(2, pmix_globals.debug_output,
                        "pmix:server constructed uri %s", myuri);

    return PMIX_SUCCESS;
}

const char* PMIx_Get_version(void)
{
    return pmix_version_string;
}

pmix_status_t PMIx_server_init(pmix_server_module_t *module,
                               int use_internal_comm)
{
    pmix_usock_posted_recv_t *req;
    int rc;
    
    ++init_cntr;
    if (1 < init_cntr) {
        return PMIX_SUCCESS;
    }

    pmix_output_verbose(2, pmix_globals.debug_output,
                        "pmix:server init called");

    /* save the comm flag for when we finalize */
    using_internal_comm = use_internal_comm;
    
    if (0 != (rc = initialize_server_base(module))) {
        return rc;
    }

    /* if the caller doesn't want us to use our own internal
     * messaging system, then we are done */
    if (!use_internal_comm) {
        return PMIX_SUCCESS;
    }
    
    /* and the usock system */
    pmix_usock_init(NULL);
    
    /* create an event base and progress thread for us */
    if (NULL == (pmix_globals.evbase = pmix_start_progress_thread())) {
        return PMIX_ERR_INIT;
    }

    /* setup the wildcard recv for inbound messages from clients */
    req = PMIX_NEW(pmix_usock_posted_recv_t);
    req->tag = UINT32_MAX;
    req->cbfunc = server_message_handler;
    /* add it to the end of the list of recvs */
    pmix_list_append(&pmix_usock_globals.posted_recvs, &req->super);

    /* start listening */
    if (PMIX_SUCCESS != pmix_start_listening(&myaddress)) {
        PMIx_server_finalize();
        return PMIX_ERR_INIT;
    }

    return PMIX_SUCCESS;
}

pmix_status_t PMIx_get_rendezvous_address(struct sockaddr_un *address, char **path)
{
    pmix_output_verbose(2, pmix_globals.debug_output,
                        "pmix:server get rendezvous address");

    memcpy(address, &myaddress, sizeof(struct sockaddr_un));
    if (NULL == path) {
        return PMIX_SUCCESS;
    }
    /* return the URI itself */
    if (NULL != myuri) {
        *path = strdup(myuri);
    } else {
        *path = NULL;
    }
    return PMIX_SUCCESS;
}

static void cleanup_server_state(void)
{
    int i;
    pmix_peer_t *peer;

    for (i=0; i < pmix_server_globals.clients.size; i++) {
        if (NULL != (peer = (pmix_peer_t*)pmix_pointer_array_get_item(&pmix_server_globals.clients, i))) {
            PMIX_RELEASE(peer);
        }
    }
    PMIX_DESTRUCT(&pmix_server_globals.clients);
    PMIX_LIST_DESTRUCT(&pmix_server_globals.nspaces);
    PMIX_LIST_DESTRUCT(&pmix_server_globals.collectives);
    PMIX_LIST_DESTRUCT(&pmix_server_globals.dmodex);
    PMIX_LIST_DESTRUCT(&pmix_server_globals.localmodex);

    if (NULL != myuri) {
        free(myuri);
    }

    pmix_bfrop_close();
    pmix_sec_finalize();
    
    pmix_output_verbose(2, pmix_globals.debug_output,
                        "pmix:server finalize complete");

    pmix_output_close(pmix_globals.debug_output);
    pmix_output_finalize();
    pmix_class_finalize();
}

pmix_status_t PMIx_server_finalize(void)
{
    if (1 != init_cntr) {
        --init_cntr;
        return PMIX_SUCCESS;
    }
    init_cntr = 0;

    pmix_output_verbose(2, pmix_globals.debug_output,
                        "pmix:server finalize called");
    
    if (using_internal_comm) {
        if (pmix_server_globals.listen_thread_active) {
            pmix_stop_listening();
        }
    
        pmix_stop_progress_thread(pmix_globals.evbase);
        event_base_free(pmix_globals.evbase);
#ifdef HAVE_LIBEVENT_GLOBAL_SHUTDOWN
        libevent_global_shutdown();
#endif
        
        if (0 <= pmix_server_globals.listen_socket) {
            CLOSE_THE_SOCKET(pmix_server_globals.listen_socket);
        }

        pmix_usock_finalize();

        /* cleanup the rendezvous file */
        unlink(myaddress.sun_path);
    }
    
    cleanup_server_state();
    pmix_output_verbose(2, pmix_globals.debug_output,
                        "pmix:server finalize complete");
    return PMIX_SUCCESS;
}

static void _register_nspace(int sd, short args, void *cbdata)
{
    pmix_setup_caddy_t *cd = (pmix_setup_caddy_t*)cbdata;
    pmix_nspace_t *nptr, *tmp;
    pmix_status_t rc;
    size_t i, j, size, rank;
    pmix_kval_t kv;
    char **nodes=NULL, **procs=NULL;
    pmix_buffer_t buf2;
    pmix_info_t *iptr;
    pmix_value_t val;
    char *msg;
    
    pmix_output_verbose(2, pmix_globals.debug_output,
                        "pmix:server _register_nspace");
    
    /* see if we already have this nspace */
    nptr = NULL;
    PMIX_LIST_FOREACH(tmp, &pmix_server_globals.nspaces, pmix_nspace_t) {
        if (0 == strcmp(tmp->nspace, cd->nspace)) {
            nptr = tmp;
            /* release any existing packed data - we will replace it */
            if (0 < nptr->server->job_info.bytes_used) {
                PMIX_DESTRUCT(&nptr->server->job_info);
                PMIX_CONSTRUCT(&nptr->server->job_info, pmix_buffer_t);
            }
            break;
        }
    }
    if (NULL == nptr) {
        nptr = PMIX_NEW(pmix_nspace_t);
        (void)strncpy(nptr->nspace, cd->nspace, PMIX_MAX_NSLEN);
        nptr->server = PMIX_NEW(pmix_server_nspace_t);
        nptr->server->nlocalprocs = cd->nlocalprocs;
        pmix_list_append(&pmix_server_globals.nspaces, &nptr->super);
    }
    /* pack the name of the nspace */
    msg = nptr->nspace;
    if (PMIX_SUCCESS != (rc = pmix_bfrop.pack(&nptr->server->job_info, &msg, 1, PMIX_STRING))) {
        PMIX_ERROR_LOG(rc);
        pmix_list_remove_item(&pmix_server_globals.nspaces, &nptr->super);
        PMIX_RELEASE(nptr);
        goto release;
    }

    /* pack the provided info */
    PMIX_CONSTRUCT(&kv, pmix_kval_t);
    for (i=0; i < cd->ninfo; i++) {
        pmix_output_verbose(2, pmix_globals.debug_output,
                            "pmix:server _register_nspace recording %s",
                            cd->info[i].key);
    
        if (0 == strcmp(cd->info[i].key, PMIX_NODE_MAP)) {
            /* parse the regex to get the argv array of node names */
            if (PMIX_SUCCESS != (rc = pmix_regex_parse_nodes(cd->info[i].value.data.string, &nodes))) {
                PMIX_ERROR_LOG(rc);
                continue;
            }
            /* if we have already found the proc map, then pass
             * the detailed map */
            if (NULL != procs) {
                pmix_pack_proc_map(&nptr->server->job_info, nodes, procs);
                pmix_argv_free(nodes);
                nodes = NULL;
                pmix_argv_free(procs);
                procs = NULL;
            }
        } else if (0 == strcmp(cd->info[i].key, PMIX_PROC_MAP)) {
            /* parse the regex to get the argv array containg proc ranks on each node */
            if (PMIX_SUCCESS != (rc = pmix_regex_parse_procs(cd->info[i].value.data.string, &procs))) {
                PMIX_ERROR_LOG(rc);
                continue;
            }
            /* if we have already recv'd the node map, then record
             * the detailed map */
            if (NULL != nodes) {
                pmix_pack_proc_map(&nptr->server->job_info, nodes, procs);
                pmix_argv_free(nodes);
                nodes = NULL;
                pmix_argv_free(procs);
                procs = NULL;
            }
        } else if (0 == strcmp(cd->info[i].key, PMIX_PROC_DATA)) {
            /* an array of data pertaining to a specific proc */
            if (PMIX_INFO_ARRAY != cd->info[i].value.type) {
                PMIX_ERROR_LOG(PMIX_ERR_BAD_PARAM);
                goto release;
            }
            size = cd->info[i].value.data.array.size;
            iptr = (pmix_info_t*)cd->info[i].value.data.array.array;
            PMIX_CONSTRUCT(&buf2, pmix_buffer_t);
            /* first element of the array must be the rank */
            if (0 != strcmp(iptr[0].key, PMIX_RANK)) {
                PMIX_ERROR_LOG(PMIX_ERR_BAD_PARAM);
                PMIX_DESTRUCT(&buf2);
                goto release;
            }
            /* pack it separately */
            rank = iptr[0].value.data.integer;
            if (PMIX_SUCCESS != (rc = pmix_bfrop.pack(&buf2, &rank, 1, PMIX_INT))) {
                PMIX_ERROR_LOG(rc);
                pmix_list_remove_item(&pmix_server_globals.nspaces, &nptr->super);
                PMIX_RELEASE(nptr);
                PMIX_DESTRUCT(&buf2);
                goto release;
            }
            /* cycle thru the values for this rank and pack them */
            for (j=1; j < size; j++) {
                kv.key = iptr[j].key;
                kv.value = &iptr[j].value;
                if (PMIX_SUCCESS != (rc = pmix_bfrop.pack(&buf2, &kv, 1, PMIX_KVAL))) {
                    PMIX_ERROR_LOG(rc);
                    pmix_list_remove_item(&pmix_server_globals.nspaces, &nptr->super);
                    PMIX_RELEASE(nptr);
                    PMIX_DESTRUCT(&buf2);
                    goto release;
                }
            }
            /* now add the blob */
            kv.key = PMIX_PROC_BLOB;
            kv.value = &val;
            val.type = PMIX_BYTE_OBJECT;
            val.data.bo.bytes = buf2.base_ptr;
            val.data.bo.size = buf2.bytes_used;
            if (PMIX_SUCCESS != (rc = pmix_bfrop.pack(&nptr->server->job_info, &kv, 1, PMIX_KVAL))) {
                PMIX_ERROR_LOG(rc);
                pmix_list_remove_item(&pmix_server_globals.nspaces, &nptr->super);
                PMIX_RELEASE(nptr);
                PMIX_DESTRUCT(&buf2);
                goto release;
            }
            PMIX_DESTRUCT(&buf2);
        } else {
            /* just a value relating to the entire job */
            kv.key = cd->info[i].key;
            kv.value = &cd->info[i].value;
            if (PMIX_SUCCESS != (rc = pmix_bfrop.pack(&nptr->server->job_info, &kv, 1, PMIX_KVAL))) {
                PMIX_ERROR_LOG(rc);
                pmix_list_remove_item(&pmix_server_globals.nspaces, &nptr->super);
                PMIX_RELEASE(nptr);
                goto release;
            }
        }
    }
    /* do not destruct the kv object - no memory leak will result */
    
 release:
    if (NULL != nodes) {
        pmix_argv_free(nodes);
    }
    if (NULL != procs) {
        pmix_argv_free(procs);
    }
    if (NULL != cd->opcbfunc) {
        cd->opcbfunc(rc, cd->cbdata);
    }
    PMIX_RELEASE(cd);
}

/* setup the data for a job */
pmix_status_t PMIx_server_register_nspace(const char nspace[], int nlocalprocs,
                                          pmix_info_t info[], size_t ninfo,
                                          pmix_op_cbfunc_t cbfunc, void *cbdata)
{
    pmix_setup_caddy_t *cd;
    size_t i;
    
    cd = PMIX_NEW(pmix_setup_caddy_t);
    (void)strncpy(cd->nspace, nspace, PMIX_MAX_NSLEN);
    cd->nlocalprocs = nlocalprocs;
    cd->opcbfunc = cbfunc;
    cd->cbdata = cbdata;
    /* copy across the info array, if given */
    if (0 < ninfo) {
        cd->ninfo = ninfo;
        PMIX_INFO_CREATE(cd->info, ninfo);
        for (i=0; i < ninfo; i++) {
            (void)strncpy(cd->info[i].key, info[i].key, PMIX_MAX_KEYLEN);
            pmix_value_xfer(&cd->info[i].value, &info[i].value);
        }
    }
    
    if (using_internal_comm) {
        /* we have to push this into our event library to avoid
         * potential threading issues */
        event_assign(&cd->ev, pmix_globals.evbase, -1,
                          EV_WRITE, _register_nspace, cd);
        event_active(&cd->ev, EV_WRITE, 1);
    } else {
        /* the caller is responsible for thread protection */
        _register_nspace(0, 0, (void*)cd);
    }
    return PMIX_SUCCESS;
}

static void _execute_collective(int sd, short args, void *cbdata)
{
    pmix_trkr_caddy_t *tcd = (pmix_trkr_caddy_t*)cbdata;
    pmix_server_trkr_t *trk = tcd->trk;
    char *data = NULL;
    size_t sz = 0;
    pmix_buffer_t bucket, pbkt, xfer;
    pmix_rank_info_t *info;
    pmix_value_t *val;
    
    /* we don't need to check for non-NULL APIs here as
     * that was already done when the tracker was created */
    if (PMIX_FENCENB_CMD == trk->type) {
        /* if the user asked us to collect data, then we have
         * to provide any locally collected data to the host
         * server so they can circulate it - only take data
         * from the specified procs as not everyone is necessarily
         * participating! And only take data intended for remote
         * distribution as local data will be added when we send
         * the result to our local clients */
        PMIX_CONSTRUCT(&bucket, pmix_buffer_t);
        if (trk->collect_data) {
            PMIX_LIST_FOREACH(info, &trk->ranks, pmix_rank_info_t) {
                PMIX_CONSTRUCT(&pbkt, pmix_buffer_t);
                /* get any remote contribution - note that there
                 * may not be a contribution */
                if (PMIX_SUCCESS == pmix_hash_fetch(&info->nptr->server->myremote, info->rank, "modex", &val) &&
                    NULL != val) {
                    PMIX_CONSTRUCT(&xfer, pmix_buffer_t);
                    PMIX_LOAD_BUFFER(&xfer, val->data.bo.bytes, val->data.bo.size);
                    pmix_buffer_t *pxfer = &xfer;
                    pmix_bfrop.pack(&pbkt, &pxfer, 1, PMIX_BUFFER);
                    xfer.base_ptr = NULL;
                    xfer.bytes_used = 0;
                    PMIX_DESTRUCT(&xfer);
                    PMIX_VALUE_RELEASE(val);
                }
                /* now pack this proc's contribution into the bucket */
                pmix_buffer_t *ppbkt = &pbkt;
                pmix_bfrop.pack(&bucket, &ppbkt, 1, PMIX_BUFFER);
                PMIX_DESTRUCT(&pbkt);
            }
        }
        PMIX_UNLOAD_BUFFER(&bucket, data, sz);
        PMIX_DESTRUCT(&bucket);
        pmix_host_server.fence_nb(trk->pcs, trk->npcs,
                                  data, sz, trk->modexcbfunc, trk);
    } else if (PMIX_CONNECTNB_CMD == trk->type) {
        pmix_host_server.connect(trk->pcs, trk->npcs,
                                 trk->op_cbfunc, trk);
    } else if (PMIX_DISCONNECTNB_CMD == trk->type) {
        pmix_host_server.disconnect(trk->pcs, trk->npcs,
                                    trk->op_cbfunc, trk);
    } else {
        /* unknown type */
        PMIX_ERROR_LOG(PMIX_ERR_NOT_FOUND);
        pmix_list_remove_item(&pmix_server_globals.collectives, &trk->super);
        PMIX_RELEASE(trk);
    }
    PMIX_RELEASE(tcd);
}

static void _register_client(int sd, short args, void *cbdata)
{
    pmix_setup_caddy_t *cd = (pmix_setup_caddy_t*)cbdata;
    pmix_rank_info_t *info;
    pmix_nspace_t *nptr, *tmp;
    pmix_server_trkr_t *trk;
    pmix_trkr_caddy_t *tcd;
    bool found;
    pmix_local_modex_caddy_t *lcd, *lcdnext;
    
    pmix_output_verbose(2, pmix_globals.debug_output,
                        "pmix:server _register_client for nspace %s rank %d",
                        cd->nspace, cd->rank);
    
    /* see if we already have this nspace */
    nptr = NULL;
    PMIX_LIST_FOREACH(tmp, &pmix_server_globals.nspaces, pmix_nspace_t) {
        if (0 == strcmp(tmp->nspace, cd->nspace)) {
            nptr = tmp;
            break;
        }
    }
    if (NULL == nptr) {
        nptr = PMIX_NEW(pmix_nspace_t);
        (void)strncpy(nptr->nspace, cd->nspace, PMIX_MAX_NSLEN);
        /* add the server object */
        nptr->server = PMIX_NEW(pmix_server_nspace_t);
        pmix_list_append(&pmix_server_globals.nspaces, &nptr->super);
    }
    /* setup a peer object for this client - since the host server
     * only deals with the original processes and not any clones,
     * we know this function will be called only once per rank */
    info = PMIX_NEW(pmix_rank_info_t);
    PMIX_RETAIN(nptr);
    info->nptr = nptr;
    info->rank = cd->rank;
    info->uid = cd->uid;
    info->gid = cd->gid;
    info->server_object = cd->server_object;
    pmix_list_append(&nptr->server->ranks, &info->super);
    /* see if we have everyone */
    if (nptr->server->nlocalprocs == pmix_list_get_size(&nptr->server->ranks)) {
        nptr->server->all_registered = true;
        /* check any pending trackers to see if they are
         * waiting for us. There is a slight race condition whereby
         * the host server could have spawned the local client and
         * it called back into the collective -before- our local event
         * would fire the register_client callback. Deal with that here. */
        PMIX_LIST_FOREACH(trk, &pmix_server_globals.collectives, pmix_server_trkr_t) {
            /* if this tracker is already complete, then we
             * don't need to update it */
            if (trk->def_complete) {
                continue;
            }
            /* is this now completed? */
            if (pmix_list_get_size(&trk->local_cbs) == trk->nlocal) {
                /* it did, so now we need to process it */
                if (using_internal_comm) {
                    /* we don't want to block someone
                     * here, so kick any completed trackers into a
                     * new event for processing */
                    PMIX_EXECUTE_COLLECTIVE(tcd, trk, _execute_collective);
                } else {
                    /* have to do it here as there is no
                     * event engine behind us */
                    PMIX_SETUP_COLLECTIVE(tcd, trk);
                    _execute_collective(0, 0, tcd);
                }
            }
        }
        /* also check any pending local modex requests to see if
         * someone has been waiting for a request on a remote proc
         * in one of our nspaces, but we didn't know all the local procs
         * and so couldn't determine the proc was remote */
        PMIX_LIST_FOREACH_SAFE(lcd, lcdnext, &pmix_server_globals.localmodex, pmix_local_modex_caddy_t) {
            if (0 != strncmp(nptr->nspace, lcd->nspace, PMIX_MAX_NSLEN)) {
                continue;
            }
            /* is this one of our local ranks? */
            found = false;
            PMIX_LIST_FOREACH(info, &nptr->server->ranks, pmix_rank_info_t) {
                if (info->rank == lcd->rank) {
                    found = true;
                    break;
                }
            }
            /* if we found it, then we don't need to do anything as we'll
             * process the request when this proc commits its modex data */
            if (!found) {
                /* remote rank - request dmodex assistance from host server */
                if (NULL == pmix_host_server.direct_modex) {
                    /* service isn't available - let the requestor know */
                    lcd->cbfunc(PMIX_ERR_NOT_FOUND, NULL, 0, lcd->cbdata);
                    continue;
                } else {
                    /* the required server caddy so the callback knows
                     * how to return the result to the requestor was
                     * included when we cached this */
                    pmix_host_server.direct_modex(lcd->nspace, lcd->rank,
                                                  lcd->cbfunc, lcd->cbdata);
                }
                pmix_list_remove_item(&pmix_server_globals.localmodex, &lcd->super);
                PMIX_RELEASE(lcd);
            }
        }
    }
    /* let the caller know we are done */
    if (NULL != cd->opcbfunc) {
        cd->opcbfunc(PMIX_SUCCESS, cd->cbdata);
    }
    PMIX_RELEASE(cd);
}

pmix_status_t PMIx_server_register_client(const char nspace[], int rank,
                                          uid_t uid, gid_t gid, void *server_object,
                                          pmix_op_cbfunc_t cbfunc, void *cbdata)
{
    pmix_setup_caddy_t *cd;

    pmix_output_verbose(2, pmix_globals.debug_output,
                        "pmix:server register client %s:%d",
                        nspace, rank);
    
     cd = PMIX_NEW(pmix_setup_caddy_t);
    (void)strncpy(cd->nspace, nspace, PMIX_MAX_NSLEN);
    cd->rank = rank;
    cd->uid = uid;
    cd->gid = gid;
    cd->server_object = server_object;
    cd->opcbfunc = cbfunc;
    cd->cbdata = cbdata;
    
    if (using_internal_comm) {
        /* we have to push this into our event library to avoid
         * potential threading issues */
        event_assign(&cd->ev, pmix_globals.evbase, -1,
                          EV_WRITE, _register_client, cd);
        event_active(&cd->ev, EV_WRITE, 1);
    } else {
        /* the caller is responsible for thread protection */
        _register_client(0, 0, (void*)cd);
    }
    return PMIX_SUCCESS;
}

/* setup the envars for a child process */
pmix_status_t PMIx_server_setup_fork(const char nspace[],
                                     int rank, char ***env)
{
    char rankstr[PMIX_MAX_VALLEN+1];
    
    pmix_output_verbose(2, pmix_globals.debug_output,
                        "pmix:server setup_fork for nspace %s rank %d",
                        nspace, rank);
    
    /* pass the nspace */
    pmix_setenv("PMIX_NAMESPACE", nspace, true, env);
    /* pass the rank */
    (void)snprintf(rankstr, PMIX_MAX_VALLEN, "%d", rank);
    pmix_setenv("PMIX_RANK", rankstr, true, env);
    /* pass our rendezvous info */
    pmix_setenv("PMIX_SERVER_URI", myuri, true, env);
    /* pass our active security mode */
    pmix_setenv("PMIX_SECURITY_MODE", security_mode, true, env);

    return PMIX_SUCCESS;
}

/***************************************************************************************************
 *  Support calls from the host server down to us requesting direct modex data provided by one     *
 *  of our local clients                                                                           *
 ***************************************************************************************************/

static void _dmodex_req(int sd, short args, void *cbdata)
{
    pmix_setup_caddy_t *cd = (pmix_setup_caddy_t*)cbdata;
    pmix_rank_info_t *info, *iptr;
    pmix_nspace_t *nptr, *ns;
    pmix_buffer_t pbkt, xfer;
    pmix_value_t *val;
    char *data;
    size_t sz;
    pmix_dmodex_caddy_t *dcd;
    
    /* this should be one of my clients, but a race condition
     * could cause this request to arrive prior to us having
     * been informed of it - so first check to see if we know
     * about this nspace yet */
    nptr = NULL;
    PMIX_LIST_FOREACH(ns, &pmix_server_globals.nspaces, pmix_nspace_t) {
        if (0 == strcmp(ns->nspace, cd->nspace)) {
            nptr = ns;
            break;
        }
    }
    if (NULL == nptr) {
        /* we don't know this namespace yet, and so we obviously
         * haven't received the data from this proc yet - defer
         * the request until we do */
        dcd = PMIX_NEW(pmix_dmodex_caddy_t);
        dcd->cd = cd;
        pmix_list_append(&pmix_server_globals.dmodex, &dcd->super);
        return;
    }

    /* see if we have this peer in our list */
    info = NULL;
    PMIX_LIST_FOREACH(iptr, &nptr->server->ranks, pmix_rank_info_t) {
        if (iptr->rank == cd->rank) {
            info = iptr;
            break;
        }
    }
    if (NULL == info) {
        /* rank isn't known yet - defer
         * the request until we do */
        dcd = PMIX_NEW(pmix_dmodex_caddy_t);
        dcd->cd = cd;
        pmix_list_append(&pmix_server_globals.dmodex, &dcd->super);
        return;
    }

    /* have we received the modex from this proc yet - if
     * not, then defer */
    if (!info->modex_recvd) {
        /* track the request so we can fulfill it once
         * data is recvd */
        dcd = PMIX_NEW(pmix_dmodex_caddy_t);
        dcd->cd = cd;
        pmix_list_append(&pmix_server_globals.dmodex, &dcd->super);
        return;
    }
    
    /* collect the remote/global data from this proc */
    PMIX_CONSTRUCT(&pbkt, pmix_buffer_t);
    /* get any remote contribution - note that there
     * may not be a contribution */
    if (PMIX_SUCCESS == pmix_hash_fetch(&nptr->server->myremote, info->rank, "modex", &val)) {
        PMIX_CONSTRUCT(&xfer, pmix_buffer_t);
        PMIX_LOAD_BUFFER(&xfer, val->data.bo.bytes, val->data.bo.size);
        pmix_buffer_t *pxfer = &xfer;
        pmix_bfrop.pack(&pbkt, &pxfer, 1, PMIX_BUFFER);
        xfer.base_ptr = NULL;
        xfer.bytes_used = 0;
        PMIX_DESTRUCT(&xfer);
        if (NULL != val) {
            PMIX_VALUE_RELEASE(val);
        }
    }
    PMIX_UNLOAD_BUFFER(&pbkt, data, sz);
    PMIX_DESTRUCT(&pbkt);

    /* execute the callback */
    cd->cbfunc(PMIX_SUCCESS, data, sz, cd->cbdata);
    if (NULL != data) {
        free(data);
    }
}

pmix_status_t PMIx_server_dmodex_request(const char nspace[], int rank,
                                         pmix_dmodex_response_fn_t cbfunc,
                                         void *cbdata)
{
    pmix_setup_caddy_t *cd;

    pmix_output_verbose(2, pmix_globals.debug_output,
                        "pmix:server register client %s:%d",
                        nspace, rank);

    /* protect against bozo */
    if (NULL == cbfunc || NULL == nspace) {
        return PMIX_ERR_BAD_PARAM;
    }
    
    cd = PMIX_NEW(pmix_setup_caddy_t);
    (void)strncpy(cd->nspace, nspace, PMIX_MAX_NSLEN);
    cd->rank = rank;
    cd->cbfunc = cbfunc;
    cd->cbdata = cbdata;
    
    if (using_internal_comm) {
        /* we have to push this into our event library to avoid
         * potential threading issues */
        event_assign(&cd->ev, pmix_globals.evbase, -1,
                     EV_WRITE, _dmodex_req, cd);
        event_active(&cd->ev, EV_WRITE, 1);
        PMIX_WAIT_FOR_COMPLETION(cd->active);
    } else {
        /* the caller is responsible for thread protection */
        _dmodex_req(0, 0, (void*)cd);
    }
    PMIX_RELEASE(cd);
    return PMIX_SUCCESS;
}


static void _notify_error(int sd, short args, void *cbdata)
{
    pmix_notify_caddy_t *cd = (pmix_notify_caddy_t*)cbdata;
    int rc;
    int i;
    size_t j;
    pmix_peer_t *peer;

    /* pack the status */
    if (PMIX_SUCCESS != (rc = pmix_bfrop.pack(cd->buf, &cd->status, 1, PMIX_INT))) {
        PMIX_ERROR_LOG(rc);
        return;
    }

    /* pack the error procs */
    if (PMIX_SUCCESS != (rc = pmix_bfrop.pack(cd->buf, &cd->error_nprocs, 1, PMIX_SIZE))) {
        PMIX_ERROR_LOG(rc);
        return;
    }
    if (0 < cd->error_nprocs) {
        if (PMIX_SUCCESS != (rc = pmix_bfrop.pack(cd->buf, cd->error_procs, cd->error_nprocs, PMIX_PROC))) {
            PMIX_ERROR_LOG(rc);
            return;
        }
    }

    /* pack the info */
    if (PMIX_SUCCESS != (rc = pmix_bfrop.pack(cd->buf, &cd->ninfo, 1, PMIX_SIZE))) {
        PMIX_ERROR_LOG(rc);
        return;
    }
    if (0 < cd->ninfo) {
        if (PMIX_SUCCESS != (rc = pmix_bfrop.pack(cd->buf, cd->info, cd->ninfo, PMIX_INFO))) {
            PMIX_ERROR_LOG(rc);
            return;
        }
    }

    if (!using_internal_comm) {
        /* it is up to the host server to decide who should get this
         * message, and to send it - so just return the data here */
        return;
    }

    /* cycle across our connected clients and send the message to
     * any within the specified proc */
    for (i=0; i < pmix_server_globals.clients.size; i++) {
        if (NULL == (peer = (pmix_peer_t*)pmix_pointer_array_get_item(&pmix_server_globals.clients, i))) {
            continue;
        }
        /* if the procs field is NULL, then send it to everyone */
        if (NULL == cd->procs) {
            PMIX_RETAIN(cd->buf);
            PMIX_SERVER_QUEUE_REPLY(peer, 0, cd->buf);
            continue;
        }
        /* otherwise, check to see if this proc is included */
        for (j=0; j < cd->nprocs; j++) {
            if (0 != strncmp(peer->info->nptr->nspace, cd->procs[j].nspace, PMIX_MAX_NSLEN)) {
                continue;
            }
            if (PMIX_RANK_WILDCARD == cd->procs[j].rank ||
                cd->procs[j].rank == peer->info->rank) {
                PMIX_RETAIN(cd->buf);
                PMIX_SERVER_QUEUE_REPLY(peer, 0, cd->buf);
            } else {
            }
        }
    }
    /* notify the caller */
    if (NULL != cd->cbfunc) {
        cd->cbfunc(PMIX_SUCCESS, cd->cbdata);
    }
    PMIX_RELEASE(cd);
}


pmix_status_t PMIx_server_notify_error(pmix_status_t status,
                                       pmix_proc_t procs[], size_t nprocs,
                                       pmix_proc_t error_procs[], size_t error_nprocs,
                                       pmix_info_t info[], size_t ninfo,
                                       char **payload, size_t *size,
                                       pmix_op_cbfunc_t cbfunc, void *cbdata)
{
    pmix_notify_caddy_t *cd;
    size_t n;
    
    cd = PMIX_NEW(pmix_notify_caddy_t);
    cd->status = status;
    /* need to copy the procs and info fields to avoid
     * any question of ownership */
    if (NULL != procs && 0 < nprocs) {
        PMIX_PROC_CREATE(cd->procs, nprocs);
        memcpy(cd->procs, procs, nprocs * sizeof(pmix_proc_t));
        cd->nprocs = nprocs;
    }
    if (NULL != error_procs && 0 < error_nprocs) {
        PMIX_PROC_CREATE(cd->error_procs, error_nprocs);
        memcpy(cd->error_procs, error_procs, error_nprocs * sizeof(pmix_proc_t));
        cd->error_nprocs = error_nprocs;
    }
    if (NULL != info && 0 < ninfo) {
        PMIX_INFO_CREATE(cd->info, ninfo);
        for (n=0; n < ninfo; n++) {
            memcpy(cd->info[n].key, info[n].key, PMIX_MAX_KEYLEN);
            pmix_value_xfer(&cd->info[n].value, &info[n].value);
        }
        cd->ninfo = ninfo;
    }
    cd->cbfunc = cbfunc;
    cd->cbdata = cbdata;
    
    if (using_internal_comm) {
        /* we have to push this into our event library to avoid
         * potential threading issues */
        event_assign(&cd->ev, pmix_globals.evbase, -1,
                          EV_WRITE, _notify_error, cd);
        event_active(&cd->ev, EV_WRITE, 1);
        return PMIX_SUCCESS;
    }

        /*set defaults */
    if (NULL == payload) {
        /* we can't process this as we have no way
         * to hand it back */
        return PMIX_ERR_BAD_PARAM;
    }
    *payload = NULL;
    *size = 0;

    /* the caller is responsible for thread protection */
    _notify_error(0, 0, (void*)cd);
    *payload = cd->buf->base_ptr;
    *size = cd->buf->bytes_used;
    /* protect the data */
    cd->buf->bytes_used = 0;
    cd->buf->base_ptr = NULL;
    PMIX_RELEASE(cd);
    return PMIX_SUCCESS;
    
}

void PMIx_Register_errhandler(pmix_notification_fn_t err)
{
    pmix_globals.errhandler = err;
}

void PMIx_Deregister_errhandler(void)
{
   pmix_globals.errhandler = NULL;
}

#define PMIX_MAX_NODE_PREFIX        50

pmix_status_t PMIx_generate_regex(const char *input, char **regexp)
{
    char *vptr, *vsave;
    char prefix[PMIX_MAX_NODE_PREFIX];
    int i, j, len, startnum, vnum, numdigits;
    bool found, fullval;
    char *suffix, *sfx;
    pmix_regex_value_t *vreg;
    pmix_regex_range_t *range;
    pmix_list_t vids;
    char **regexargs = NULL, *tmp, *tmp2;
    char *cptr;

    /* define the default */
    *regexp = NULL;

    /* setup the list of results */
    PMIX_CONSTRUCT(&vids, pmix_list_t);

    /* cycle thru the array of input values - first copy
     * it so we don't overwrite what we were given*/
    vsave = strdup(input);
    vptr = vsave;
    while (NULL != (cptr = strchr(vptr, ',')) || 0 < strlen(vptr)) {
        if (NULL != cptr) {
            *cptr = '\0';
        }
        /* determine this node's prefix by looking for first non-alpha char */
        fullval = false;
        len = strlen(vptr);
        startnum = -1;
        memset(prefix, 0, PMIX_MAX_NODE_PREFIX);
        numdigits = 0;
        for (i=0, j=0; i < len; i++) {
            if (!isalpha(vptr[i])) {
                /* found a non-alpha char */
                if (!isdigit(vptr[i])) {
                    /* if it is anything but a digit, we just use
                     * the entire name
                     */
                    fullval = true;
                    break;
                }
                /* count the size of the numeric field - but don't
                 * add the digits to the prefix
                 */
                numdigits++;
                if (startnum < 0) {
                    /* okay, this defines end of the prefix */
                    startnum = i;
                }
                continue;
            }
            if (startnum < 0) {
                prefix[j++] = vptr[i];
            }
        }
        if (fullval || startnum < 0) {
            /* can't compress this name - just add it to the list */
            vreg = PMIX_NEW(pmix_regex_value_t);
            vreg->prefix = strdup(vptr);
            pmix_list_append(&vids, &vreg->super);
            /* move to the next posn */
            if (NULL == cptr) {
                break;
            }
            vptr = cptr + 1;
            continue;
        }
        /* convert the digits and get any suffix */
        vnum = strtol(&vptr[startnum], &sfx, 10);
        if (NULL != sfx) {
            suffix = strdup(sfx);
        } else {
            suffix = NULL;
        }
        /* is this value already on our list? */
        found = false;
        PMIX_LIST_FOREACH(vreg, &vids, pmix_regex_value_t) {
            if (0 < strlen(prefix) && NULL == vreg->prefix) {
                continue;
            }
            if (0 == strlen(prefix) && NULL != vreg->prefix) {
                continue;
            }
            if (0 < strlen(prefix) && NULL != vreg->prefix
                && 0 != strcmp(prefix, vreg->prefix)) {
                continue;
            }
            if (NULL == suffix && NULL != vreg->suffix) {
                continue;
            }
            if (NULL != suffix && NULL == vreg->suffix) {
                continue;
            }
            if (NULL != suffix && NULL != vreg->suffix &&
                0 != strcmp(suffix, vreg->suffix)) {
                continue;
            }
            if (numdigits != vreg->num_digits) {
                continue;
            }
            /* found a match - flag it */
            found = true;
            /* get the last range on this nodeid - we do this
             * to preserve order
             */
            range = (pmix_regex_range_t*)pmix_list_get_last(&vreg->ranges);
            if (NULL == range) {
                /* first range for this value */
                range = PMIX_NEW(pmix_regex_range_t);
                range->start = vnum;
                range->cnt = 1;
                pmix_list_append(&vreg->ranges, &range->super);
                break;
            }
            /* see if the value is out of sequence */
            if (vnum != (range->start + range->cnt)) {
                /* start a new range */
                range = PMIX_NEW(pmix_regex_range_t);
                range->start = vnum;
                range->cnt = 1;
                pmix_list_append(&vreg->ranges, &range->super);
                break;
            }
            /* everything matches - just increment the cnt */
            range->cnt++;
            break;
        }
        if (!found) {
            /* need to add it */
            vreg = PMIX_NEW(pmix_regex_value_t);
            if (0 < strlen(prefix)) {
                vreg->prefix = strdup(prefix);
            }
            if (NULL != suffix) {
                vreg->suffix = strdup(suffix);
            }
            vreg->num_digits = numdigits;
            pmix_list_append(&vids, &vreg->super);
            /* record the first range for this value - we took
             * care of values we can't compress above
             */
            range = PMIX_NEW(pmix_regex_range_t);
            range->start = vnum;
            range->cnt = 1;
            pmix_list_append(&vreg->ranges, &range->super);
        }
        if (NULL != suffix) {
            free(suffix);
        }
        /* move to the next posn */
        if (NULL == cptr) {
            break;
        }
        vptr = cptr + 1;
    }
    free(vsave);

    /* begin constructing the regular expression */
    while (NULL != (vreg = (pmix_regex_value_t*)pmix_list_remove_first(&vids))) {        
        /* if no ranges, then just add the name */
        if (0 == pmix_list_get_size(&vreg->ranges)) {
            if (NULL != vreg->prefix) {
                /* solitary value */
                asprintf(&tmp, "%s", vreg->prefix);
                pmix_argv_append_nosize(&regexargs, tmp);
                free(tmp);
            }
            PMIX_RELEASE(vreg);
            continue;
        }
        /* start the regex for this value with the prefix */
        if (NULL != vreg->prefix) {
            asprintf(&tmp, "%s[%d:", vreg->prefix, vreg->num_digits);
        } else {
            asprintf(&tmp, "[%d:", vreg->num_digits);
        }
        /* add the ranges */
        while (NULL != (range = (pmix_regex_range_t*)pmix_list_remove_first(&vreg->ranges))) {
            if (1 == range->cnt) {
                asprintf(&tmp2, "%s%d,", tmp, range->start);
            } else {
                asprintf(&tmp2, "%s%d-%d,", tmp, range->start, range->start + range->cnt - 1);
            }
            free(tmp);
            tmp = tmp2;
            PMIX_RELEASE(range);
        }
        /* replace the final comma */
        tmp[strlen(tmp)-1] = ']';
        if (NULL != vreg->suffix) {
            /* add in the suffix, if provided */
            asprintf(&tmp2, "%s%s", tmp, vreg->suffix);
            free(tmp);
            tmp = tmp2;
        }
        pmix_argv_append_nosize(&regexargs, tmp);
        free(tmp);
        PMIX_RELEASE(vreg);
    }
    
    /* assemble final result */
    tmp = pmix_argv_join(regexargs, ',');
    asprintf(regexp, "pmix[%s]", tmp);
    free(tmp);
    
    /* cleanup */
    pmix_argv_free(regexargs);

    PMIX_DESTRUCT(&vids);
    return PMIX_SUCCESS;
}

pmix_status_t PMIx_generate_ppn(const char *input, char **regexp)
{
    char **ppn, **npn;
    int i, j, start, end;
    pmix_regex_value_t *vreg;
    pmix_regex_range_t *rng;
    pmix_list_t nodes;
    char *tmp, *tmp2;
    char *cptr;

    /* define the default */
    *regexp = NULL;

    /* setup the list of results */
    PMIX_CONSTRUCT(&nodes, pmix_list_t);

    /* split the input by node */
    ppn = pmix_argv_split(input, ';');

    /* for each node, split the input by comma */
    for (i=0; NULL != ppn[i]; i++) {
        rng = NULL;
        /* create a record for this node */
        vreg = PMIX_NEW(pmix_regex_value_t);
        pmix_list_append(&nodes, &vreg->super);
        /* split the input for this node */
        npn = pmix_argv_split(ppn[i], ',');
        /* look at each element */
        for (j=0; NULL != npn[j]; j++) {
            /* is this a range? */
            if (NULL != (cptr = strchr(npn[j], '-'))) {
                /* terminate the string */
                *cptr = '\0';
                ++cptr;
                start = strtol(npn[j], NULL, 10);
                end = strtol(cptr, NULL, 10);
                /* are we collecting a range? */
                if (NULL == rng) {
                    /* no - better start one */
                    rng = PMIX_NEW(pmix_regex_range_t);
                    rng->start = start;
                    rng->cnt = end - start + 1;
                    pmix_list_append(&vreg->ranges, &rng->super);
                } else {
                    /* is this a continuation of the current range? */
                    if (start == (rng->start + rng->cnt)) {
                        /* just add it to the end of this range */
                        rng->cnt++;
                    } else {
                        /* nope, there is a break - create new range */
                        rng = PMIX_NEW(pmix_regex_range_t);
                        rng->start = start;
                        rng->cnt = end - start + 1;
                        pmix_list_append(&vreg->ranges, &rng->super);
                    }
                }
            } else {
                /* single rank given */
                start = strtol(npn[j], NULL, 10);
                /* are we collecting a range? */
                if (NULL == rng) {
                    /* no - better start one */
                    rng = PMIX_NEW(pmix_regex_range_t);
                    rng->start = start;
                    rng->cnt = 1;
                    pmix_list_append(&vreg->ranges, &rng->super);
                } else {
                    /* is this a continuation of the current range? */
                    if (start == (rng->start + rng->cnt)) {
                        /* just add it to the end of this range */
                        rng->cnt++;
                    } else {
                        /* nope, there is a break - create new range */
                        rng = PMIX_NEW(pmix_regex_range_t);
                        rng->start = start;
                        rng->cnt = 1;
                        pmix_list_append(&vreg->ranges, &rng->super);
                    }
                }
            }
        }
        pmix_argv_free(npn);
    }
    pmix_argv_free(ppn);
    

    /* begin constructing the regular expression */
    tmp = strdup("pmix[");
    PMIX_LIST_FOREACH(vreg, &nodes, pmix_regex_value_t) {
        while (NULL != (rng = (pmix_regex_range_t*)pmix_list_remove_first(&vreg->ranges))) {
            if (1 == rng->cnt) {
                asprintf(&tmp2, "%s%d,", tmp, rng->start);
            } else {
                asprintf(&tmp2, "%s%d-%d,", tmp, rng->start, rng->start + rng->cnt - 1);
            }
            free(tmp);
            tmp = tmp2;
            PMIX_RELEASE(rng);
        }
        /* replace the final comma */
        tmp[strlen(tmp)-1] = ';';
    }
    
    /* replace the final semi-colon */
    tmp[strlen(tmp)-1] = ']';
    
    /* assemble final result */
    *regexp = tmp;

    PMIX_LIST_DESTRUCT(&nodes);
    return PMIX_SUCCESS;
}

/****    THE FOLLOWING CALLBACK FUNCTIONS ARE USED BY THE HOST SERVER    ****
 ****    THEY THEREFORE CAN OCCUR IN EITHER THE HOST SERVER'S THREAD     ****
 ****    CONTEXT, OR IN OUR OWN THREAD CONTEXT IF THE CALLBACK OCCURS    ****
 ****    IMMEDIATELY. THUS ANYTHING THAT ACCESSES A GLOBAL ENTITY        ****
 ****    MUST BE PUSHED INTO AN EVENT FOR PROTECTION                     ****/

static void op_cbfunc(int status, void *cbdata)
{
    pmix_server_caddy_t *cd = (pmix_server_caddy_t*)cbdata;
    pmix_buffer_t *reply;
    int rc;

    /* setup the reply with the returned status */
    reply = PMIX_NEW(pmix_buffer_t);
    if (PMIX_SUCCESS != (rc = pmix_bfrop.pack(reply, &status, 1, PMIX_INT))) {
        PMIX_ERROR_LOG(rc);
        PMIX_RELEASE(reply);
        return;
    }
    /* the function that created the server_caddy did a
     * retain on the peer, so we don't have to worry about
     * it still being present - send a copy to the originator */
    PMIX_SERVER_QUEUE_REPLY(cd->peer, cd->hdr.tag, reply);
    /* cleanup */
    PMIX_RELEASE(cd);
}

static void spawn_cbfunc(int status, char *nspace, void *cbdata)
{
    pmix_server_caddy_t *cd = (pmix_server_caddy_t*)cbdata;
    pmix_buffer_t *reply;
    int rc;
    pmix_nspace_t *nptr, *ns;
    
    pmix_output(0, "SPAWN_CBFUNC: %d %s", status, nspace);
    /* setup the reply with the returned status */
    reply = PMIX_NEW(pmix_buffer_t);
    if (PMIX_SUCCESS != (rc = pmix_bfrop.pack(reply, &status, 1, PMIX_INT))) {
        PMIX_ERROR_LOG(rc);
        PMIX_RELEASE(reply);
        return;
    }
    /* add any job-related info we have on that nspace - this will
     * include the name of the nspace */
    nptr = NULL;
    PMIX_LIST_FOREACH(ns, &pmix_server_globals.nspaces, pmix_nspace_t) {
        if (0 == strcmp(ns->nspace, nspace)) {
            nptr = ns;
            break;
        }
    }
    if (NULL == nptr) {
        /* shouldn't happen */
        PMIX_ERROR_LOG(PMIX_ERR_NOT_FOUND);
    } else {
        pmix_bfrop.copy_payload(reply, &nptr->server->job_info);
    }
    /* the function that created the server_caddy did a
     * retain on the peer, so we don't have to worry about
     * it still being present - tell the originator the result */
    PMIX_SERVER_QUEUE_REPLY(cd->peer, cd->hdr.tag, reply);
    /* cleanup */
    PMIX_RELEASE(cd);
}

static void lookup_cbfunc(int status, pmix_pdata_t pdata[], size_t ndata,
                          void *cbdata)
{
    pmix_server_caddy_t *cd = (pmix_server_caddy_t*)cbdata;
    pmix_buffer_t *reply;
    int rc;

    /* setup the reply with the returned status */
    reply = PMIX_NEW(pmix_buffer_t);
    if (PMIX_SUCCESS != (rc = pmix_bfrop.pack(reply, &status, 1, PMIX_INT))) {
        PMIX_ERROR_LOG(rc);
        PMIX_RELEASE(reply);
        return;
    }
    if (PMIX_SUCCESS == status) {
        /* pack the returned data objects */
        if (PMIX_SUCCESS != (rc = pmix_bfrop.pack(reply, &ndata, 1, PMIX_SIZE))) {
            PMIX_ERROR_LOG(rc);
            PMIX_RELEASE(reply);
            return;
        }
        if (PMIX_SUCCESS != (rc = pmix_bfrop.pack(reply, pdata, ndata, PMIX_PDATA))) {
            PMIX_ERROR_LOG(rc);
            PMIX_RELEASE(reply);
            return;
        }
    }

    /* the function that created the server_caddy did a
     * retain on the peer, so we don't have to worry about
     * it still being present - tell the originator the result */
    PMIX_SERVER_QUEUE_REPLY(cd->peer, cd->hdr.tag, reply);
    /* cleanup */
    PMIX_RELEASE(cd);
}

static void _tracker_complete(int sd, short args, void *cbdata)
{
    pmix_trkr_caddy_t *cd = (pmix_trkr_caddy_t*)cbdata;
    
    pmix_list_remove_item(&pmix_server_globals.collectives, &cd->trk->super);
    PMIX_RELEASE(cd->trk);
    PMIX_RELEASE(cd);
}
    
static void modex_cbfunc(int status, const char *data,
                         size_t ndata, void *cbdata)
{
    pmix_server_trkr_t *tracker = (pmix_server_trkr_t*)cbdata;
    pmix_buffer_t *reply, xfer, pbkt;
    int rc;
    pmix_server_caddy_t *cd;
    pmix_rank_info_t *info;
    pmix_value_t *val;
    
    pmix_output_verbose(2, pmix_globals.debug_output,
                        "server:modex_cbfunc called with %d bytes", (int)ndata);

    if (NULL == tracker) {
        /* nothing to do */
        return;
    }
    
    /* setup the reply, starting with the returned status */
    reply = PMIX_NEW(pmix_buffer_t);
    if (PMIX_SUCCESS != (rc = pmix_bfrop.pack(reply, &status, 1, PMIX_INT))) {
        PMIX_ERROR_LOG(rc);
        PMIX_RELEASE(reply);
        return;
    }
    /* pass the blobs being returned */
    PMIX_CONSTRUCT(&xfer, pmix_buffer_t);
    PMIX_LOAD_BUFFER(&xfer, data, ndata);
    pmix_bfrop.copy_payload(reply, &xfer);
    /* protect the incoming data */
    xfer.base_ptr = NULL;
    xfer.bytes_used = 0;
    /* cleanup */
    PMIX_DESTRUCT(&xfer);

    /* if collect_data is set, then find all local procs involved in
     * the modex, and add their "local" modex data as well - it
     * would not have been included when we sent this out */
    if (tracker->collect_data) {
        PMIX_LIST_FOREACH(info, &tracker->ranks, pmix_rank_info_t) {
            PMIX_CONSTRUCT(&pbkt, pmix_buffer_t);
            /* get any local contribution - note that there
             * may not be a contribution */
            if (PMIX_SUCCESS == pmix_hash_fetch(&info->nptr->server->mylocal, info->rank, "modex", &val) &&
                NULL != val) {
                /* pack the proc so we know the source */
                char *foobar = info->nptr->nspace;
                pmix_bfrop.pack(&pbkt, &foobar, 1, PMIX_STRING);
                pmix_bfrop.pack(&pbkt, &info->rank, 1, PMIX_INT);
                PMIX_CONSTRUCT(&xfer, pmix_buffer_t);
                PMIX_LOAD_BUFFER(&xfer, val->data.bo.bytes, val->data.bo.size);
                pmix_buffer_t *pxfer = &xfer;
                pmix_bfrop.pack(&pbkt, &pxfer, 1, PMIX_BUFFER);
                xfer.base_ptr = NULL;
                xfer.bytes_used = 0;
                PMIX_DESTRUCT(&xfer);
                PMIX_VALUE_RELEASE(val);
                /* now pack this proc's contribution into the bucket */
                pmix_buffer_t *ppbkt = &pbkt;
                pmix_bfrop.pack(reply, &ppbkt, 1, PMIX_BUFFER);
            }
            PMIX_DESTRUCT(&pbkt);
        }
    }

    /* loop across all procs in the tracker, sending them the reply */
    PMIX_LIST_FOREACH(cd, &tracker->local_cbs, pmix_server_caddy_t) {
        PMIX_RETAIN(reply);
        pmix_output_verbose(2, pmix_globals.debug_output,
                            "server:modex_cbfunc reply being sent to %s:%d",
                            cd->peer->info->nptr->nspace, cd->peer->info->rank);
        PMIX_SERVER_QUEUE_REPLY(cd->peer, cd->hdr.tag, reply);
    }
    PMIX_RELEASE(reply);  // maintain accounting
    /* push this into our own thread so we can remove it from the
     * global collectives */
    PMIX_MARK_COLLECTIVE_COMPLETE(tracker, _tracker_complete);
}

static void get_cbfunc(int status, const char *data,
                       size_t ndata, void *cbdata)
{
    pmix_server_caddy_t *cd = (pmix_server_caddy_t*)cbdata;
    pmix_buffer_t *reply, buf;
    int rc;
    
    pmix_output_verbose(2, pmix_globals.debug_output,
                        "server:get_cbfunc called with %d elements", (int)ndata);

    if (NULL == cd) {
        /* nothing to do */
        return;
    }
    
    /* setup the reply, starting with the returned status */
    reply = PMIX_NEW(pmix_buffer_t);
    if (PMIX_SUCCESS != (rc = pmix_bfrop.pack(reply, &status, 1, PMIX_INT))) {
        PMIX_ERROR_LOG(rc);
        goto cleanup;
    }
    /* pack the blob being returned */
    PMIX_CONSTRUCT(&buf, pmix_buffer_t);
    PMIX_LOAD_BUFFER(&buf, data, ndata);
    pmix_bfrop.copy_payload(reply, &buf);
    buf.base_ptr = NULL;
    buf.bytes_used = 0;
    PMIX_DESTRUCT(&buf);
    /* send the data to the requestor */
    pmix_output_verbose(2, pmix_globals.debug_output,
                        "server:get_cbfunc reply being sent to %s:%d",
                        cd->peer->info->nptr->nspace, cd->peer->info->rank);
    PMIX_SERVER_QUEUE_REPLY(cd->peer, cd->hdr.tag, reply);

 cleanup:
    PMIX_RELEASE(cd);
}

static void cnct_cbfunc(int status, void *cbdata)
{
    pmix_server_trkr_t *tracker = (pmix_server_trkr_t*)cbdata;
    pmix_buffer_t *reply;
    int rc, i;
    pmix_server_caddy_t *cd;
    char **nspaces=NULL;
    pmix_nspace_t *nptr;
    pmix_buffer_t *job_info_ptr;
    
    pmix_output_verbose(2, pmix_globals.debug_output,
                        "server:cnct_cbfunc called");

    if (NULL == tracker) {
        /* nothing to do */
        return;
    }
    
    /* setup the reply, starting with the returned status */
    reply = PMIX_NEW(pmix_buffer_t);
    if (PMIX_SUCCESS != (rc = pmix_bfrop.pack(reply, &status, 1, PMIX_INT))) {
        PMIX_ERROR_LOG(rc);
        PMIX_RELEASE(reply);
        return;
    }

    if (PMIX_CONNECTNB_CMD == tracker->type) {
        /* find the unique nspaces that are participating */
        PMIX_LIST_FOREACH(cd, &tracker->local_cbs, pmix_server_caddy_t) {
            pmix_argv_append_unique_nosize(&nspaces, cd->peer->info->nptr->nspace, false);
        }
    
        /* loop across all participating nspaces and include their
         * job-related info */
        for (i=0; NULL != nspaces[i]; i++) {
            PMIX_LIST_FOREACH(nptr, &pmix_server_globals.nspaces, pmix_nspace_t) {
                if (0 != strcmp(nspaces[i], nptr->nspace)) {
                    continue;
                }
                job_info_ptr = &nptr->server->job_info;
                if (PMIX_SUCCESS != (rc = pmix_bfrop.pack(reply, &job_info_ptr, 1, PMIX_BUFFER))) {
                    PMIX_ERROR_LOG(rc);
                    PMIX_RELEASE(reply);
                    pmix_argv_free(nspaces);
                    return;
                }
            }
        }
        pmix_argv_free(nspaces);
    }
    
    /* loop across all procs in the tracker, sending them the reply */
    PMIX_LIST_FOREACH(cd, &tracker->local_cbs, pmix_server_caddy_t) {
        PMIX_RETAIN(reply);
        pmix_output_verbose(2, pmix_globals.debug_output,
                            "server:cnct_cbfunc reply being sent to %s:%d",
                            cd->peer->info->nptr->nspace, cd->peer->info->rank);
        PMIX_SERVER_QUEUE_REPLY(cd->peer, cd->hdr.tag, reply);
    }
    PMIX_RELEASE(reply);  // maintain accounting
    /* push this into our own thread so we can remove it from the
     * global collectives */
    PMIX_MARK_COLLECTIVE_COMPLETE(tracker, _tracker_complete);
}


/* the switchyard is the primary message handling function. It's purpose
 * is to take incoming commands (packed into a buffer), unpack them,
 * and then call the corresponding host server's function to execute
 * them. Some commands involve only a single proc (i.e., the one
 * sending the command) and can be executed while we wait. In these cases,
 * the switchyard will construct and pack a reply buffer to be returned
 * to the sender.
 *
 * Other cases (either multi-process collective or cmds that require
 * an async reply) cannot generate an immediate reply. In these cases,
 * the reply buffer will be NULL. An appropriate callback function will
 * be called that will be responsible for eventually replying to the
 * calling processes.
 *
 * Should an error be encountered at any time within the switchyard, an
 * error reply buffer will be returned so that the caller can be notified,
 * thereby preventing the process from hanging. */ 
static int server_switchyard(pmix_peer_t *peer, uint32_t tag,
                             pmix_buffer_t *buf)
{
    int rc;
    int32_t cnt;
    pmix_cmd_t cmd;
    pmix_server_caddy_t *cd;
    
    /* retrieve the cmd */
    cnt = 1;
    if (PMIX_SUCCESS != (rc = pmix_bfrop.unpack(buf, &cmd, &cnt, PMIX_CMD))) {
        PMIX_ERROR_LOG(rc);
        return rc;
    }
    pmix_output_verbose(2, pmix_globals.debug_output,
                        "recvd pmix cmd %d from %s:%d",
                        cmd, peer->info->nptr->nspace, peer->info->rank);

    if (PMIX_ABORT_CMD == cmd) {
        PMIX_PEER_CADDY(cd, peer, tag); 
        if (PMIX_SUCCESS != (rc = pmix_server_abort(peer, buf, op_cbfunc, cd))) {
            PMIX_ERROR_LOG(rc);
            PMIX_RELEASE(cd);
        }
        return rc;
    }
        
    if (PMIX_COMMIT_CMD == cmd) {
        if (PMIX_SUCCESS != (rc = pmix_server_commit(peer, buf))) {
            PMIX_ERROR_LOG(rc);
        }
        return rc;
    }
        
    if (PMIX_FENCENB_CMD == cmd) {
        PMIX_PEER_CADDY(cd, peer, tag);
        if (PMIX_SUCCESS != (rc = pmix_server_fence(cd, buf, modex_cbfunc, op_cbfunc))) {
            PMIX_ERROR_LOG(rc);
            PMIX_RELEASE(cd);
        }
        return rc;
    }

    if (PMIX_GETNB_CMD == cmd) {
        PMIX_PEER_CADDY(cd, peer, tag);
        if (PMIX_SUCCESS != (rc = pmix_server_get(buf, get_cbfunc, cd))) {
            PMIX_ERROR_LOG(rc);
            PMIX_RELEASE(cd);
        }
        return rc;
    }
        
    if (PMIX_FINALIZE_CMD == cmd) {
        pmix_output_verbose(2, pmix_globals.debug_output,
                            "recvd FINALIZE");
        /* call the local server, if supported */
        if (NULL != pmix_host_server.finalized) {
            PMIX_PEER_CADDY(cd, peer, tag);
            if (PMIX_SUCCESS != (rc = pmix_host_server.finalized(peer->info->nptr->nspace,
                                                                 peer->info->rank,
                                                                 peer->info->server_object,
                                                                 op_cbfunc, cd))) {
                PMIX_ERROR_LOG(rc);
                PMIX_RELEASE(cd);
            }
        }
        /* turn off the recv event - we shouldn't hear anything
         * more from this proc */
        if (peer->recv_ev_active) {
            event_del(&peer->recv_event);
            peer->recv_ev_active = false;
        }
        return rc;
    }

        
    if (PMIX_PUBLISHNB_CMD == cmd) {
        PMIX_PEER_CADDY(cd, peer, tag);
        if (PMIX_SUCCESS != (rc = pmix_server_publish(peer, buf, op_cbfunc, cd))) {
            PMIX_ERROR_LOG(rc);
            PMIX_RELEASE(cd);
        }
        return rc;
    }

    
    if (PMIX_LOOKUPNB_CMD == cmd) {
        PMIX_PEER_CADDY(cd, peer, tag);
        if (PMIX_SUCCESS != (rc = pmix_server_lookup(buf, lookup_cbfunc, cd))) {
            PMIX_ERROR_LOG(rc);
            PMIX_RELEASE(cd);
        }
        return rc;
    }

        
    if (PMIX_UNPUBLISHNB_CMD == cmd) {
        PMIX_PEER_CADDY(cd, peer, tag);
        if (PMIX_SUCCESS != (rc = pmix_server_unpublish(buf, op_cbfunc, cd))) {
            PMIX_ERROR_LOG(rc);
            PMIX_RELEASE(cd);
        }
        return rc;
    }

        
    if (PMIX_SPAWNNB_CMD == cmd) {
        PMIX_PEER_CADDY(cd, peer, tag);
        if (PMIX_SUCCESS != (rc = pmix_server_spawn(buf, spawn_cbfunc, cd))) {
            PMIX_ERROR_LOG(rc);
            PMIX_RELEASE(cd);
        }
        return rc;
    }

        
    if (PMIX_CONNECTNB_CMD == cmd) {
        PMIX_PEER_CADDY(cd, peer, tag);
        if (PMIX_SUCCESS != (rc = pmix_server_connect(cd, buf, false, cnct_cbfunc))) {
            PMIX_ERROR_LOG(rc);
            PMIX_RELEASE(cd);
        }
        return rc;
    }

    if (PMIX_DISCONNECTNB_CMD == cmd) {
        PMIX_PEER_CADDY(cd, peer, tag);
        if (PMIX_SUCCESS != (rc = pmix_server_connect(cd, buf, true, cnct_cbfunc))) {
            PMIX_ERROR_LOG(rc);
            PMIX_RELEASE(cd);
        }
        return rc;
    }

    return PMIX_ERR_NOT_SUPPORTED;
}

static void server_message_handler(struct pmix_peer_t *pr, pmix_usock_hdr_t *hdr,
                                   pmix_buffer_t *buf, void *cbdata)
{
    pmix_peer_t *peer = (pmix_peer_t*)pr;
    pmix_buffer_t *reply;
    int rc;

    pmix_output_verbose(2, pmix_globals.debug_output,
                        "SWITCHYARD for %s:%d:%d",
                        peer->info->nptr->nspace,
                        peer->info->rank, peer->sd);

    rc = server_switchyard(peer, hdr->tag, buf);
    /* send the return, if there was an error returned */
    if (PMIX_SUCCESS != rc) {
        reply = PMIX_NEW(pmix_buffer_t);
        pmix_bfrop.pack(reply, &rc, 1, PMIX_INT);
        PMIX_SERVER_QUEUE_REPLY(peer, hdr->tag, reply);
        PMIX_RELEASE(reply);
    }
}
